
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  ALLOC.C
*   DESCRIP  :  FENRIS Operating System Abstraction Layer
*   DATE     :  November 1, 1998
*
*
***************************************************************************/

#include "globals.h"

// we track our memory to check for leaks at unload.

typedef struct _MEM_HEADER
{
   struct _MEM_HEADER *next;
   struct _MEM_HEADER *prior;
   ULONG Size;
   ULONG Type;
   ULONG Pool;
   void *context;
} MEM_HEADER;

MEM_HEADER *MemoryListHead = 0;
MEM_HEADER *MemoryListTail = 0;

ULONG AddToMemoryList(MEM_HEADER *mem);
ULONG RemoveFromMemoryList(MEM_HEADER *mem);


#if (LINUX_24)

#define toupper(c) (((c) >= 'a' && (c) <= 'z') ? (c) + ('A'-'a') : (c))
#define tolower(c) (((c) >= 'A' && (c) <= 'Z') ? (c)-('A'-'a') : (c))

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    sema_init(sema, value);
    return 0;
}

long WaitOnSemaphore(void *sema)
{
#if (DEBUG_DEADLOCKS)
    register ULONG ccode;

    ccode = down_interruptible(sema);
    if (ccode == -EINTR)
       return ccode;
    return 0;
#else
    down(sema);
    return 0;
#endif
}

long SignalSemaphore(void *sema)
{
    up(sema);
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_KERNEL);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;
   mem->context = 0;
   
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_BUFFER);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 3;
   mem->context = 0;
   
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register ULONG nSize;
   register MEM_HEADER *mem;
   register BYTE *p;

   nSize = (((size + sizeof(MEM_HEADER)) + 4095) / 4096) * 4096;
   mem = vmalloc(nSize);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 2;
   mem->context = 0;
   
   PagedMemoryAllocated += size;
   PagedMemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
    register MEM_HEADER *mem, *mem1;

    mem1 = p;
    mem = &mem1[-1];
    switch (mem->Pool)
    {
       case 1:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
         MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      kfree((void *)mem);
      break;

       case 2:
      PagedMemoryFreed += mem->Size;
      if (PagedMemoryInUse)
         PagedMemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      vfree((void *)mem);
      break;

       case 3:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
         MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      kfree((void *)mem);
      break;

       default:
      NWFSPrint("nwfs:  NWFSFree reports free on invalid memory\n");
      break;
    }
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
    copy_from_user(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
    copy_to_user(dest, src, size);
}

void NWFSSetUserSpace(void *dest, ULONG value, ULONG size)
{
    if (size > 65536)
       return;

    if (ZeroBuffer)
       copy_to_user(dest, ZeroBuffer, size);
}

BYTE NWFSUpperCase(BYTE c)
{
    return (toupper(c));
}

BYTE NWFSLowerCase(BYTE c)
{
    return (tolower(c));
}

void NWFSClean(void *p)
{
   kfree((void *)p);
}

#endif


#if (LINUX_22)

#define toupper(c) (((c) >= 'a' && (c) <= 'z') ? (c) + ('A'-'a') : (c))
#define tolower(c) (((c) >= 'A' && (c) <= 'Z') ? (c)-('A'-'a') : (c))

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    struct semaphore *semaphore = sema;

    NWFSSet(semaphore, 0, sizeof(struct semaphore));
    semaphore->count.counter = value;
    return 0;
}

long WaitOnSemaphore(void *sema)
{
#if (DEBUG_DEADLOCKS)
    register ULONG ccode;

    ccode = down_interruptible(sema);
    if (ccode == -EINTR)
       return ccode;
    return 0;
#else
    down(sema);
    return 0;
#endif
}

long SignalSemaphore(void *sema)
{
    up(sema);
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_KERNEL);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_BUFFER);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register ULONG nSize;
   register MEM_HEADER *mem;
   register BYTE *p;

   nSize = (((size + sizeof(MEM_HEADER)) + 4095) / 4096) * 4096;
   mem = vmalloc(nSize);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 2;

   PagedMemoryAllocated += size;
   PagedMemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
    register MEM_HEADER *mem, *mem1;

    mem1 = p;
    mem = &mem1[-1];
    switch (mem->Pool)
    {
       case 1:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
         MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      kfree((void *)mem);
      break;

       case 2:
      PagedMemoryFreed += mem->Size;
      if (PagedMemoryInUse)
         PagedMemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      vfree((void *)mem);
      break;

       default:
      NWFSPrint("nwfs:  NWFSFree reports free on invalid memory\n");
      break;
    }
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
    copy_from_user(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
    copy_to_user(dest, src, size);
}

void NWFSSetUserSpace(void *dest, ULONG value, ULONG size)
{
    if (size > 65536)
       return;

    if (ZeroBuffer)
       copy_to_user(dest, ZeroBuffer, size);
}

BYTE NWFSUpperCase(BYTE c)
{
    return (toupper(c));
}

BYTE NWFSLowerCase(BYTE c)
{
    return (tolower(c));
}

void NWFSClean(void *p)
{
   kfree((void *)p);
}

#endif


#if (LINUX_20)

#define toupper(c) (((c) >= 'a' && (c) <= 'z') ? (c) + ('A'-'a') : (c))
#define tolower(c) (((c) >= 'A' && (c) <= 'Z') ? (c)-('A'-'a') : (c))

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    struct semaphore *semaphore = sema;

    NWFSSet(semaphore, 0, sizeof(struct semaphore));
    semaphore->count = value;
    return 0;
}

long WaitOnSemaphore(void *sema)
{
#if (DEBUG_DEADLOCKS)
    register ULONG ccode;

    ccode = down_interruptible(sema);
    if (ccode == -EINTR)
       return ccode;
    return 0;
#else
    down(sema);
    return 0;
#endif
}

long SignalSemaphore(void *sema)
{
    up(sema);
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_KERNEL);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = kmalloc((size + sizeof(MEM_HEADER)), GFP_BUFFER);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register ULONG nSize;
   register MEM_HEADER *mem;
   register BYTE *p;

   nSize = (((size + sizeof(MEM_HEADER)) + 4095) / 4096) * 4096;
   mem = vmalloc(nSize);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 2;

   PagedMemoryAllocated += size;
   PagedMemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
    register MEM_HEADER *mem, *mem1;

    mem1 = p;
    mem = &mem1[-1];
    switch (mem->Pool)
    {
       case 1:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
	 MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      kfree((void *)mem);
      break;

       case 2:
      PagedMemoryFreed += mem->Size;
      if (PagedMemoryInUse)
         PagedMemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      vfree((void *)mem);
      break;

       default:
      NWFSPrint("nwfs:  NWFSFree reports free on invalid memory\n");
      break;
    }
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
    memcpy_fromfs(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
    memcpy_tofs(dest, src, size);
}

void NWFSSetUserSpace(void *dest, ULONG value, ULONG size)
{
    if (size > 65536)
       size = 65536;

    if (ZeroBuffer)
       memcpy_tofs(dest, ZeroBuffer, size);
}

BYTE NWFSUpperCase(BYTE c)
{
    return (toupper(c));
}

BYTE NWFSLowerCase(BYTE c)
{
    return (tolower(c));
}

void NWFSClean(void *p)
{
   kfree((void *)p);
}

#endif



#if (LINUX_UTIL)

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    return 0;
}

long WaitOnSemaphore(void *sema)
{
    return 0;
}

long SignalSemaphore(void *sema)
{
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc((size + sizeof(MEM_HEADER)));
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc((size + sizeof(MEM_HEADER)));
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register ULONG nSize;
   register MEM_HEADER *mem;
   register BYTE *p;

   nSize = (((size + sizeof(MEM_HEADER)) + 4095) / 4096) * 4096;
   mem = malloc(nSize);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 2;

   PagedMemoryAllocated += size;
   PagedMemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
    register MEM_HEADER *mem, *mem1;

    mem1 = p;
    mem = &mem1[-1];
    switch (mem->Pool)
    {
       case 1:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
         MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      free((void *)mem);
      break;

       case 2:
      PagedMemoryFreed += mem->Size;
      if (PagedMemoryInUse)
         PagedMemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      free((void *)mem);
      break;

       default:
      NWFSPrint("nwfs:  NWFSFree reports free on invalid memory\n");
      break;
    }
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
    memcpy(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
    memcpy(dest, src, size);
}

void NWFSSetUserSpace(void *src, ULONG value, ULONG size)
{
    memset(src, value, size);
}

BYTE NWFSUpperCase(BYTE c)
{
    return (toupper(c));
}

BYTE NWFSLowerCase(BYTE c)
{
    return (tolower(c));
}

void NWFSClean(void *p)
{
   free((void *)p);
}

#endif



#if (WINDOWS_NT | WINDOWS_NT_RO)

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    return 0;
}

long WaitOnSemaphore(void *sema)
{
    return 0;
}

long SignalSemaphore(void *sema)
{
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = ExAllocatePoolWithTag(PagedPool,
                   size + sizeof(MEM_HEADER),
                   NT_POOL_TAG);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = ExAllocatePoolWithTag(PagedPoolCacheAligned,
                   size + sizeof(MEM_HEADER),
                   NT_POOL_TAG);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = ExAllocatePoolWithTag(PagedPoolCacheAligned,
                   size + sizeof(MEM_HEADER),
                   NT_POOL_TAG);
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
   register MEM_HEADER *mem, *mem1;

   mem1 = p;
   mem = &mem1[-1];
   MemoryFreed += mem->Size;
   if (MemoryInUse)
      MemoryInUse -= mem->Size;

   RemoveFromMemoryList(mem);

   ExFreePool((void *)mem);
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
   memmove(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
   memmove(dest, src, size);
}

void NWFSSetUserSpace(void *dest, ULONG value, ULONG size)
{
   memset(dest, value, size);
}

void NWFSClean(void *p)
{
   ExFreePool((void *)p);
}

#endif



#if (WINDOWS_NT_UTIL)

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    return 0;
}

long WaitOnSemaphore(void *sema)
{
    return 0;
}

long SignalSemaphore(void *sema)
{
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc(size + sizeof(MEM_HEADER));
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc(size + sizeof(MEM_HEADER));
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc(size + sizeof(MEM_HEADER));
   if (!mem)
      return 0;

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
   register MEM_HEADER *mem, *mem1;

   mem1 = p;
   mem = &mem1[-1];
   MemoryFreed += mem->Size;
   if (MemoryInUse)
      MemoryInUse -= mem->Size;

   RemoveFromMemoryList(mem);

   free((void *)mem);
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
   memmove(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
   memmove(dest, src, size);
}

void NWFSSetUserSpace(void *dest, ULONG value, ULONG size)
{
   memset(dest, value, size);
}

void NWFSClean(void *p)
{
   free((void *)p);
}

#endif


#if (DOS_UTIL)

ULONG AllocateSemaphore(void *sema, ULONG value)
{
    return 0;
}

long WaitOnSemaphore(void *sema)
{
    return 0;
}

long SignalSemaphore(void *sema)
{
    return 0;
}

void *NWFSAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc((size + sizeof(MEM_HEADER)));
   if (!mem)
   {
      NWFSPrint("alloc failed [%-4s]\n", (char *)Tag ? (char *)Tag : "NONE");
      return 0;
   }

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSIOAlloc(ULONG size, ULONG Tag)
{
   register MEM_HEADER *mem;
   register BYTE *p;

   mem = malloc((size + sizeof(MEM_HEADER)));
   if (!mem)
   {
      NWFSPrint("alloc failed [%-4s]\n", (char *)Tag ? (char *)Tag : "NONE");
      return 0;
   }

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 1;

   MemoryAllocated += size;
   MemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;

}

void *NWFSCacheAlloc(ULONG size, ULONG Tag)
{
   register ULONG nSize;
   register MEM_HEADER *mem;
   register BYTE *p;

   nSize = (((size + sizeof(MEM_HEADER)) + 4095) / 4096) * 4096;
   mem = malloc(nSize);
   if (!mem)
   {
      NWFSPrint("cache alloc failed [%-4s]\n", (char *)Tag ? (char *)Tag : "NONE");
      return 0;
   }

   p = (BYTE *)((ULONG)mem + sizeof(MEM_HEADER));
   mem->Size = size;
   mem->Type = Tag;
   mem->next = 0;
   mem->prior = 0;
   mem->Pool = 2;

   PagedMemoryAllocated += size;
   PagedMemoryInUse += size;

   AddToMemoryList(mem);

   return (void *) p;
}

void NWFSFree(void *p)
{
    register MEM_HEADER *mem, *mem1;

    mem1 = p;
    mem = &mem1[-1];
    switch (mem->Pool)
    {
       case 1:
      MemoryFreed += mem->Size;
      if (MemoryInUse)
         MemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      free((void *)mem);
      break;

       case 2:
      PagedMemoryFreed += mem->Size;
      if (PagedMemoryInUse)
         PagedMemoryInUse -= mem->Size;
      RemoveFromMemoryList(mem);
      free((void *)mem);
      break;

       default:
      NWFSPrint("nwfs:  NWFSFree reports free on invalid memory\n");
      break;
    }
}

int NWFSCompare(void *dest, void *src, ULONG size)
{
    return (memcmp(dest, src, size));
}

void NWFSSet(void *dest, ULONG value, ULONG size)
{
    memset(dest, value, size);
}

void NWFSCopy(void *dest, void *src, ULONG size)
{
    memmove(dest, src, size);
}

void NWFSCopyFromUserSpace(void *dest, void *src, ULONG size)
{
    memcpy(dest, src, size);
}

void NWFSCopyToUserSpace(void *dest, void *src, ULONG size)
{
    memcpy(dest, src, size);
}

void NWFSSetUserSpace(void *src, ULONG value, ULONG size)
{
    memset(src, value, size);
}

BYTE NWFSUpperCase(BYTE c)
{
    return (toupper(c));
}

BYTE NWFSLowerCase(BYTE c)
{
    return (tolower(c));
}

void NWFSClean(void *p)
{
   free((void *)p);
}

#endif

void displayMemoryList(void)
{
    BYTE *p;
    MEM_HEADER *mem, *temp;

    mem = MemoryListHead;
    while (mem)
    {
       p = (BYTE *) &mem->Type;
       NWFSPrint("addr-%08X type-[%s] size-%d\n",
         (unsigned int)mem, (char *)mem->Type ? (char *)mem->Type : "NONE",
         (int)mem->Size);
       temp = mem;
       mem = mem->next;
       NWFSClean(temp);  // free the tracked memory
    }
    return;
}

ULONG AddToMemoryList(MEM_HEADER *mem)
{
    if (!MemoryListHead)
    {
       MemoryListHead = mem;
       MemoryListTail = mem;
       mem->next = mem->prior = 0;
    }
    else
    {
       MemoryListTail->next = mem;
       mem->next = 0;
       mem->prior = MemoryListTail;
       MemoryListTail = mem;
    }
    return 0;
}

ULONG RemoveFromMemoryList(MEM_HEADER *mem)
{
    if (MemoryListHead == mem)
    {
       MemoryListHead = mem->next;
       if (MemoryListHead)
      MemoryListHead->prior = NULL;
       else
      MemoryListTail = NULL;
    }
    else
    {
       mem->prior->next = mem->next;
       if (mem != MemoryListTail)
      mem->next->prior = mem->prior;
       else
      MemoryListTail = mem->prior;
    }
    return 0;
}



